@@ -1,4 +1,4 @@
-/* gok-output.h
+/* gok-output.c
 *
 * Copyright 2001,2002 Sun Microsystems, Inc.,
 * Copyright 2001,2002 University Of Toronto
@@ -31,28 +31,87 @@
 #include "gok-word-complete.h"
 #include "main.h"
 
-gunichar gok_output_unichar_from_keysym (KeySym keysym)
+static gboolean
+gok_output_can_predict (void)
+{
+	GokKeyboard *kbd = gok_main_get_current_keyboard ();
+	return (kbd && !gok_main_get_login() && gok_data_get_wordcomplete () &&
+		gok_keyboard_get_supports_wordcomplete (kbd));
+}
+
+/**
+ * gok_output_update_predictions:
+ * Send new data to the predictor based on output type, and update the keyboard's prediction keys. 
+ **/
+static void
+gok_output_update_predictions (GokOutput    *output,
+			       gpointer      data)
+{
+	gunichar letter;
+	gchar *string;
+	gchar **prediction_list = NULL;
+
+	if (output) {
+		switch (output->Type) 
+		{
+		case OUTPUT_KEYCODE:
+		case OUTPUT_KEYSYM:
+			prediction_list = gok_wordcomplete_process_and_predict (gok_wordcomplete_get_default (),
+										(gunichar) data,
+										gok_data_get_num_predictions ());
+			gok_keyboard_set_predictions (gok_main_get_current_keyboard (),
+						      prediction_list);
+			break;
+		case OUTPUT_KEYSTRING:
+			string = (gchar *) data;
+			if (string && g_utf8_validate (string, -1, NULL)) 
+			{
+			        letter = g_utf8_get_char (string);
+				while (letter && string) 
+				{
+					if (prediction_list)
+					{
+						int i;
+						for (i = 0; prediction_list && prediction_list[i]; ++i)
+							g_free (prediction_list[i]);
+						g_free (prediction_list);
+					}
+
+					prediction_list = gok_wordcomplete_process_and_predict (
+						gok_wordcomplete_get_default (),
+						letter,
+						gok_data_get_num_predictions ());
+					string = g_utf8_next_char (string);
+					letter = g_utf8_get_char (string);
+				}
+				gok_keyboard_set_predictions (gok_main_get_current_keyboard (), 
+							      prediction_list);
+			}
+			break;
+		default:
+			break;
+		}
+	}
+}
+
+static KeySym last_dead_key = 0;
+
+gunichar gok_output_unichar_from_keysym (KeySym keysym, guint mods)
 {
 	char *charstring;
-	if (keysym <= 0xff) 
-		return (gunichar) keysym;
-	else if ((keysym >= 0xFF08) && (keysym < 0xFF20)) { /* X11 control codes, etc. */
+	if ((keysym >= 0xFF08) && (keysym < 0xFF20)) { /* X11 control codes, etc. */
+		last_dead_key = 0;
 		return (gunichar) keysym & 0x00FF; /* mask out the high bits */
 	}
+	else if ((keysym >= 0xFE50) && (keysym <= 0xFE62)) { /* deadkey: save to apply later */
+		last_dead_key = keysym;
+		return 0;
+	}
 	else {
-		char buf[20];
-		int nbytes, extra_rtn;
-		Display *display;
-		if (gok_main_get_main_window ()) 
+		long ucs;
+		if ((ucs = keysym2ucs (keysym)) > 0) 
 		{
-			display = GDK_WINDOW_XDISPLAY (gok_main_get_main_window ()->window);
-			/* hmm, how do we know XkbTranslateKeySym will use UTF-8 ? */
-			/* TODO: pass current modmask instead of 0 (edge cases) */
-			nbytes = MIN (XkbTranslateKeySym (display,
-							  &keysym, 0, buf, 19, &extra_rtn), 19); 
-			if (nbytes) {
-				return g_utf8_get_char_validated (buf, nbytes);
-			}
+			return (gunichar) ucs;
 		}
 		return 0;
 	}
@@ -170,15 +229,75 @@ GokOutput* gok_output_new_from_xml (xmlNode* pNode)
 		typeOutput = OUTPUT_EXEC;
 		flagOutput = SPI_KEY_PRESSRELEASE;
 	}
+	else if (xmlStrcmp (pStringAttributeValue, (const xmlChar *)"internal")
+		== 0)
+	{
+		typeOutput = OUTPUT_INTERNAL;
+		flagOutput = SPI_KEY_PRESSRELEASE;  /* not necessary? */
+	}
 	else
 	{
 		gok_log_x ("Output type '%s' not correct!\n", pStringAttributeValue);
 		return NULL;
 	}
-	
-	return gok_output_new (typeOutput, (gchar *) xmlNodeGetContent (pNode), flagOutput);
+
+	return gok_output_new (typeOutput, g_strstrip (xmlNodeGetContent (pNode)), flagOutput);
 }
 
+static gunichar
+gok_output_unichar_postfix_from_deadkey (KeySym deadkey)
+{
+	switch (deadkey) 
+	{
+	case XK_dead_grave:
+		return 0x0300;
+	case XK_dead_acute:
+		return 0x0301;
+	case XK_dead_circumflex:
+		return 0x0302;
+	case XK_dead_tilde:
+		return 0x0303;
+	case XK_dead_macron:
+		return 0x0304;
+	case XK_dead_breve:
+		return 0x0306;
+	case XK_dead_abovedot:
+		return 0x0307;
+	case XK_dead_diaeresis:
+		return 0x0308;
+	case XK_dead_abovering:
+		return 0x030a;
+	case XK_dead_doubleacute:
+		return 0x030b;
+	case XK_dead_caron:
+		return 0x030c;
+	case XK_dead_cedilla:
+		return 0x0327;
+	case XK_dead_ogonek:
+		return 0x0328;
+	case XK_dead_iota:
+		return 0x0345;
+	case XK_dead_voiced_sound:
+		return 0x032c;
+	case XK_dead_semivoiced_sound:
+		return 0x0324;
+	case XK_dead_belowdot:
+		return 0x0323;
+/* ifdefs needed because XSun doesn't include these keysyms */
+#ifdef XK_dead_hook
+	case XK_dead_hook:
+		return 0x0309;
+#endif
+#ifdef XK_dead_horn
+	case XK_dead_horn:
+		return 0x031b;
+#endif
+	default:
+		return 0;
+	}
+}
+
+
 /**
 * gok_output_send_to_system
 * @pOutput: Pointer to a GokOutput that will be sent to the system.
@@ -190,6 +309,7 @@ void gok_output_send_to_system (GokOutput* pOutput, gboolean bWordCompletion)
 {
 	long keysym;
 	long keycode;
+	gunichar unichar;
 
 	gok_log_enter();
 	while (pOutput != NULL)
@@ -197,7 +317,7 @@ void gok_output_send_to_system (GokOutput* pOutput, gboolean bWordCompletion)
 		if (pOutput->Type == OUTPUT_KEYSYM)
 		{
 			keysym = XStringToKeysym (pOutput->Name);
-
+			
 			gok_log ("Sending a keysym");
 			gok_log ("pOutput->Name = %s", pOutput->Name);
 			gok_log ("SPI_KEY_SYM = %d", (int) SPI_KEY_SYM);
@@ -205,13 +325,25 @@ void gok_output_send_to_system (GokOutput* pOutput, gboolean bWordCompletion)
 			SPI_generateKeyboardEvent ((long) keysym, NULL, SPI_KEY_SYM);
 
 			/* update the word prediction keys with the new output */
-			if (bWordCompletion == TRUE)
-			{
-				gok_wordcomplete_predict (gok_output_unichar_from_keysym (keysym));
+			if (bWordCompletion && gok_output_can_predict ()) {
+				unichar = gok_output_unichar_from_keysym (keysym, 
+									  gok_spy_get_modmask ());
+				if (last_dead_key && unichar) {
+					gchar *s;
+					s = gok_wordcomplete_process_unichar (gok_wordcomplete_get_default (), unichar);
+					gok_output_update_predictions (pOutput, 
+								       (gpointer) gok_output_unichar_postfix_from_deadkey (last_dead_key));
+					last_dead_key = 0;
+				}
+				else if (unichar) {
+					gok_output_update_predictions (pOutput, 
+								       (gpointer) unichar);
+				}
 			}
 		}
 		else if (pOutput->Type == OUTPUT_KEYCODE)
 		{
+			XkbDescRec *xkb = gok_keyboard_get_xkb_desc ();
 			keycode = strtol (pOutput->Name, NULL, 0);
 
 			gok_log ("Sending a keycode");
@@ -219,22 +351,37 @@ void gok_output_send_to_system (GokOutput* pOutput, gboolean bWordCompletion)
 			gok_log ("pOutput->Flag = %d", (int) pOutput->Flag);
 
 			SPI_generateKeyboardEvent ((long) keycode, NULL, pOutput->Flag);
+
 			/* update the word prediction keys with the new output */
-			if (bWordCompletion == TRUE && 
+			if (bWordCompletion && gok_output_can_predict () &&
 			    !gok_key_modifier_for_keycode (
 				    gok_main_display (),
-				    gok_keyboard_get_xkb_desc (),
+				    xkb,
 				    keycode)) 
 			{
-				gulong mask = gok_spy_get_modmask ();
-			        int index = (mask & ShiftMask) ?
-					((mask & LockMask) ? 0 : 1) : 
-					((mask & LockMask) ? 1 : 0);
+				guint mask = gok_spy_get_modmask ();
+				gint group = gok_key_get_effective_group ();
+				int type = gok_key_get_xkb_type_index (xkb, keycode, group);
+			        int level = gok_key_level_for_type (gok_keyboard_get_display (),
+								    xkb, 
+								    type, &mask);
 
-				/* TODO: get the correct index from XKB */
-			        keysym = XKeycodeToKeysym (gok_main_display (), 
-							   keycode, index);
-				gok_wordcomplete_predict (gok_output_unichar_from_keysym (keysym));
+			        keysym = XkbKeycodeToKeysym (gok_main_display (), 
+							   keycode, group, level);
+				
+				unichar = gok_output_unichar_from_keysym (keysym, 
+									  gok_spy_get_modmask ());
+				if (last_dead_key && unichar) {
+					gchar *s;
+					s = gok_wordcomplete_process_unichar (gok_wordcomplete_get_default (), unichar);
+					gok_output_update_predictions (pOutput, 
+								       (gpointer) gok_output_unichar_postfix_from_deadkey (last_dead_key));
+					last_dead_key = 0;
+				}
+				else if (unichar) {
+					gok_output_update_predictions (pOutput, 
+								       (gpointer) unichar);
+				}
 			}
 		}
 		else if (pOutput->Type == OUTPUT_KEYSTRING)
@@ -245,10 +392,10 @@ void gok_output_send_to_system (GokOutput* pOutput, gboolean bWordCompletion)
 			SPI_generateKeyboardEvent (0, pOutput->Name, SPI_KEY_STRING);
 
 			/* update the word prediction keys with the new output */
-			if (bWordCompletion == TRUE)
+			if (bWordCompletion == TRUE && gok_output_can_predict ())
 			{
-				/* TODO: allow word completion with substring insertion */
-				gok_wordcomplete_predict (0);
+				gok_output_update_predictions (pOutput,
+							       pOutput->Name);
 			}
 		}
 		else if (pOutput->Type == OUTPUT_EXEC)
@@ -258,6 +405,11 @@ void gok_output_send_to_system (GokOutput* pOutput, gboolean bWordCompletion)
 
 			g_spawn_command_line_async (pOutput->Name, NULL);
 		}
+		else if (pOutput->Type == OUTPUT_INTERNAL)
+		{
+			gok_log ("Internal command: %s", pOutput->Name);
+			gok_output_internal( pOutput->Name );
+		}
 		else
 		{
 			gok_log_x ("Output has invalid type '%d'!\n", pOutput->Type);
@@ -266,3 +418,35 @@ void gok_output_send_to_system (GokOutput* pOutput, gboolean bWordCompletion)
 	}
 	gok_log_leave();
 }
+
+/**
+ *
+ * Returns keycode for OUTPUT_KEYCODE and OUTPUT_KEYSYM keys, or -1.
+ **/
+int 
+gok_output_get_keycode (GokOutput *output)
+{
+    int keycode = -1;
+    if (output && output->Type == OUTPUT_KEYCODE)
+	keycode = atoi (output->Name);
+    else if (output && output->Type == OUTPUT_KEYSYM)
+	keycode = XKeysymToKeycode (gok_main_display (), 
+				    XStringToKeysym (output->Name));
+
+    return keycode;
+}
+
+
+void
+gok_output_internal (gchar* action)
+{
+	if (strcmp (action, "quit") == 0)
+	{
+		/* quit gok */
+		gok_main_close();
+	}
+	else
+	{
+		gok_log_x("Unknown gok internal action: %s", action);
+	}
+}
